

program SolveCosmology
        use IniFile
        use MonteCarlo
        use ParamDef
        use settings
        use cmbdata
        use posthoc
        use WeakLen
        use CalcLike
        implicit none
      
        
! This is a driving routine that illustrates the use of the program.
        character(LEN=80) InputFile, LogFile


        logical bad
        integer numsets, i, numtoget, action
        character(LEN=80) rootname,filename(100), numstr
        Type(ParamSet) Params
        integer num_points
#ifdef MPI
        double precision intime
        integer ierror

 call mpi_init(ierror)
 if (ierror/=MPI_SUCCESS) stop 'MPI fail: rank'
#endif


        InputFile = GetParam(1)
        if (InputFile == '') call DoStop('No parameter input file')
        numstr = GetParam(2)
        if (numstr /= '') then
         read(numstr,*) instance
         rand_inst = instance   
        else
         instance = 0
        end if


        call Ini_Open(InputFile, 1, bad, .false.)
        if (bad) call DoStop('Error opening parameter file')
        Ini_fail_on_not_found = .false.


#ifdef MPI 
        
        if (instance /= 0) call DoStop('With MPI should not have second parameter')
        call mpi_comm_rank(mpi_comm_world,MPIrank,ierror)

        instance = MPIrank +1 !start at 1 for chains
        write (numstr,*) instance
        rand_inst = instance
        if (ierror/=MPI_SUCCESS) call DoStop('MPI fail')

        call mpi_comm_size(mpi_comm_world,MPIchains,ierror)
        if (.not. FileExists('bessel_cache.camb')) then
         !Wait for first chain to have made the bessels file
         !this is a lazy kludge..

           call MPI_BARRIER(MPI_COMM_WORLD, ierror)
           intime = MPI_WTIME()        
           do while (MPIrank/=0 .and. MPI_WTIME()-intime < 10)
           end do           
        end if       
        
        MPI_StartTime = MPI_WTime()
        MPI_R_Stop = Ini_Read_Real('MPI_Converge_Stop',MPI_R_Stop)
        MPI_LearnPropose = Ini_Read_Logical('MPI_LearnPropose',.true.)  

        MPI_Check_Limit_Converge = Ini_Read_Logical('MPI_Check_Limit_Converge',.false.)
        if (MPI_Check_Limit_Converge) then
         MPI_Limit_Converge = Ini_Read_Real('MPI_Limit_Converge',0.025)
         MPI_Limit_Converge_Err = Ini_Read_Real('MPI_Limit_Converge_Err',0.3)
         MPI_Limit_Param = Ini_Read_Int('MPI_Limit_Param',0)
        end if
      
#endif

        Ini_fail_on_not_found = .true.
        
        rootname = Ini_Read_String('file_root')
        
        FileChangeIniAll = trim(rootname)//'.read'

        if (instance /= 0) then
            rootname = trim(rootname)//'_'//trim(adjustl(numstr))
        end if

        action = Ini_Read_Int('action',0)

        if (action ==1) call ReadPostParams

        FeedBack = Ini_Read_Int('feedback',0)
        FileChangeIni = trim(rootname)//'.read'

        if (action == 0) then
        LogFile = trim(rootname)//'.log'

        if (LogFile /= '') then
         logfile_unit = 49
         open(unit=logfile_unit,file=LogFile,form='formatted',status='replace')
         
        else
         logfile_unit = 0
        end if

        outfile_unit = 48
        open(unit=outfile_unit,file=trim(rootname)//'.txt',form='formatted',status='replace')

        indep_sample = Ini_Read_Int('indep_sample')
        if (indep_sample /=0) then
          indepfile_unit = 47
          open(unit=indepfile_unit,file=trim(rootname)//'.data',form='unformatted',status='replace')
        end if
 
        Ini_fail_on_not_found = .false.
        burn_in = Ini_Read_Int('burn_in',0)     

        end if

        numstr = Ini_Read_String('rand_seed')
        if (numstr /= '') then
          read(numstr,*) i
          call InitRandom(i)
        else
          call InitRandom()
        end if


        w_is_w = Ini_Read_Logical ('w_is_w',.true.)     
        oversample_fast = Ini_Read_Int('oversample_fast',1)
        
        if (Ini_Read_Logical('cmb_hyperparameters', .false.)) &
            call DoStop( 'Hyperparameters not supported any more')

        Use_2dF = Ini_Read_Logical('use_2dF',.false.)
        Use_Clusters = Ini_Read_Logical('use_clusters',.false.)

        Use_HST = Ini_Read_Logical('use_HST',.true.)
        Use_BBN = Ini_Read_Logical('use_BBN',.false.)
        Use_Age_Tophat_Prior= Ini_Read_Logical('use_Age_Tophat_Prior',.true.)
        Use_SN = Ini_Read_Logical('use_SN',.false.)
        Use_CMB = Ini_Read_Logical('use_CMB',.true.)
        Use_WeakLen = Ini_Read_Logical('use_WeakLen',.false.)
        Use_min_zre = Ini_Read_Double('use_min_zre',0.d0) 

        use_LSS = Use_2dF .or. Use_Clusters .or. Use_WeakLen

        Temperature = Ini_Read_Real('temperature',1.)
        
        num_threads = Ini_Read_Int('num_threads',0)
        !$ if (num_threads /=0) call OMP_SET_NUM_THREADS(num_threads)

        Ini_fail_on_not_found = .true.

        call Initialize(Params)

        
        numsets = Ini_Read_Int('cmb_numdatasets')
        do i= 1, numsets
         filename(i) = Ini_Read_String(numcat('cmb_dataset',i)) 
         !Need to close this ini file before reading in from another
        end do
        
 
        numtoget = Ini_Read_Int('samples')

        call Ini_Close

        num_points = 0
        if (Use_CMB) then
         do i= 1, numsets
        
          call ReadDataset(filename(i))
          num_points = num_points + datasets(i)%num_points
         end do
         if (Feedback > 1) write (*,*) 'read datasets'
        end if
    

        if (action == 0) then
         if (Feedback > 0) write (*,*) 'starting Monte-Carlo'
         call MCMCSample(Params, numtoget)
 
         if (Feedback > 0) write (*,*) 'finished'

         if (logfile_unit /=0) close(logfile_unit)
         if (indepfile_unit /=0) close(indepfile_unit)
        
         close(outfile_unit)

        else if (action==1) then
          if (Feedback > 0) write (*,*) 'starting post processing'
          call postprocess(rootname)
        else
         call DoStop('undefined action')
        end if

        call DoStop

end program SolveCosmology
